/*
  spamdyke -- a filter for stopping spam at connection time.
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <pwd.h>
#include <time.h>
#include <ctype.h>
#include "spamdyke.h"
#include "usage.h"
#include "dns.h"
#include "environment.h"
#include "tls.h"
#include "log.h"
#include "configuration.h"

extern int opterr;

int config_test_file_read(struct filter_settings *current_settings, char *target_file, char *option_name, char *start_message, char *success_message, char *failure_message)
  {
  int return_value;
  FILE *tmp_file;

  return_value = 1;

  if (target_file != NULL)
    {
    spamdyke_log(current_settings, LOG_LEVEL_DEBUG, start_message, option_name, target_file);

    if ((tmp_file = fopen(target_file, "r")) != NULL)
      {
      spamdyke_log(current_settings, LOG_LEVEL_DEBUG, success_message, option_name, target_file);
      fclose(tmp_file);
      }
    else
      {
      spamdyke_log(current_settings, LOG_LEVEL_ERROR, failure_message, option_name, target_file, strerror(errno));
      return_value = 0;
      }
    }

  return(return_value);
  }

int config_test_file_write(struct filter_settings *current_settings, char *target_file, char *option_name, char *start_message, char *success_message, char *failure_message)
  {
  int return_value;
  FILE *tmp_file;

  return_value = 1;

  if (target_file != NULL)
    {
    spamdyke_log(current_settings, LOG_LEVEL_DEBUG, start_message, option_name, target_file);

    if ((tmp_file = fopen(target_file, "a")) != NULL)
      {
      spamdyke_log(current_settings, LOG_LEVEL_DEBUG, success_message, option_name, target_file);
      fclose(tmp_file);
      }
    else
      {
      spamdyke_log(current_settings, LOG_LEVEL_ERROR, failure_message, option_name, target_file, strerror(errno));
      return_value = 0;
      }
    }

  return(return_value);
  }

int config_test_file_read_write(struct filter_settings *current_settings, char *target_file, char *option_name, char *start_message, char *success_message, char *failure_message)
  {
  int return_value;
  FILE *tmp_file;

  return_value = 1;

  if (target_file != NULL)
    {
    spamdyke_log(current_settings, LOG_LEVEL_DEBUG, start_message, option_name, target_file);

    if ((tmp_file = fopen(target_file, "a+")) != NULL)
      {
      spamdyke_log(current_settings, LOG_LEVEL_DEBUG, success_message, option_name, target_file);
      fclose(tmp_file);
      }
    else
      {
      spamdyke_log(current_settings, LOG_LEVEL_ERROR, failure_message, option_name, target_file, strerror(errno));
      return_value = 0;
      }
    }

  return(return_value);
  }

int config_test_file_execute(struct filter_settings *current_settings, char *target_file, char *option_name, char *start_message, char *success_message, char *failure_message)
  {
  int return_value;
  int i;
  struct stat tmp_stat;
  int num_groups;
  gid_t *tmp_gid;
  int group_match;

  return_value = 1;

  if (target_file != NULL)
    {
    spamdyke_log(current_settings, LOG_LEVEL_DEBUG, start_message, option_name, target_file);

    if (stat(target_file, &tmp_stat) == 0)
      {
      if ((tmp_stat.st_mode & S_IFMT) != S_IFREG)
        {
        spamdyke_log(current_settings, LOG_LEVEL_ERROR, failure_message, option_name, target_file, CONFIG_TEST_MSG_REGULAR_FILE);
        return_value = 0;
        }
      else
        {
        if (tmp_stat.st_uid == geteuid())
          {
          /* User owns the file. */
          if (tmp_stat.st_mode & S_IXUSR)
            spamdyke_log(current_settings, LOG_LEVEL_DEBUG, success_message, option_name, target_file);
          else
            {
            spamdyke_log(current_settings, LOG_LEVEL_ERROR, failure_message, option_name, target_file, CONFIG_TEST_MSG_OWNER_NO_EXEC);
            return_value = 0;
            }
          }
        else if (tmp_stat.st_gid == getegid())
          {
          /* Primary group owns the file. */
          if (tmp_stat.st_mode & S_IXGRP)
            spamdyke_log(current_settings, LOG_LEVEL_DEBUG, success_message, option_name, target_file);
          else
            {
            spamdyke_log(current_settings, LOG_LEVEL_ERROR, failure_message, option_name, target_file, CONFIG_TEST_MSG_PGROUP_NO_EXEC);
            return_value = 0;
            }
          }
        else
          {
          group_match = 0;

          if ((num_groups = getgroups(0, NULL)) > 0)
            {
            if ((tmp_gid = (gid_t *)malloc(sizeof(gid_t) * num_groups)) != NULL)
              {
              if (getgroups(num_groups, tmp_gid) > 0)
                for (i = 0; i < num_groups; i++)
                  if (tmp_stat.st_gid == tmp_gid[i])
                    {
                    /* Secondary group owns the file. */
                    if (tmp_stat.st_mode & S_IXGRP)
                      spamdyke_log(current_settings, LOG_LEVEL_DEBUG, success_message, option_name, target_file);
                    else
                      {
                      spamdyke_log(current_settings, LOG_LEVEL_ERROR, failure_message, option_name, target_file, CONFIG_TEST_MSG_SGROUP_NO_EXEC);
                      return_value = 0;
                      }

                    group_match = 1;
                    break;
                    }

              free(tmp_gid);
              }
            else
              {
              spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_REALLOC, (long)(sizeof(gid_t) * num_groups));
              return_value = 0;
              }
            }

          if (!group_match)
            {
            if (tmp_stat.st_mode & S_IXOTH)
              spamdyke_log(current_settings, LOG_LEVEL_DEBUG, success_message, option_name, target_file);
            else
              {
              spamdyke_log(current_settings, LOG_LEVEL_ERROR, failure_message, option_name, target_file, CONFIG_TEST_MSG_OTHER_NO_EXEC);
              return_value = 0;
              }
            }
          }
        }
      }
    else
      {
      spamdyke_log(current_settings, LOG_LEVEL_ERROR, failure_message, option_name, target_file, strerror(errno));
      return_value = 0;
      }
    }

  return(return_value);
  }

int config_test_dir_read(struct filter_settings *current_settings, char *target_dir, char *option_name, char *start_message, char *success_message, char *failure_message)
  {
  int return_value;
  int test_return;
  DIR *tmp_dir;
  struct dirent *tmp_ent;
  struct stat tmp_stat;
  char tmp_name[MAX_BUF + 1];

  return_value = 1;

  if (target_dir != NULL)
    {
    spamdyke_log(current_settings, LOG_LEVEL_DEBUG, start_message, option_name, target_dir);

    if ((tmp_dir = opendir(target_dir)) != NULL)
      {
      while ((tmp_ent = readdir(tmp_dir)) != NULL)
        {
        snprintf(tmp_name, MAX_BUF, "%s/%s", target_dir, tmp_ent->d_name);
        switch (tmp_ent->d_type)
          {
          case DT_DIR:
            if ((strcmp(tmp_ent->d_name, DIR_CURRENT) != 0) &&
                (strcmp(tmp_ent->d_name, DIR_PARENT) != 0))
              {
              test_return = config_test_dir_read(current_settings, tmp_name, option_name, NULL, NULL, failure_message);
              if (return_value)
                return_value = test_return;
              }

            break;
          case DT_REG:
            if (stat(tmp_name, &tmp_stat) != 0)
              {
              spamdyke_log(current_settings, LOG_LEVEL_ERROR, failure_message, option_name, tmp_name, strerror(errno));
              return_value = 0;
              }

            break;
          }
        }

      closedir(tmp_dir);

      if (return_value &&
          (success_message != NULL))
        spamdyke_log(current_settings, LOG_LEVEL_DEBUG, success_message, option_name, target_dir);
      }
    else
      {
      spamdyke_log(current_settings, LOG_LEVEL_ERROR, failure_message, option_name, target_dir, strerror(errno));
      return_value = 0;
      }
    }

  return(return_value);
  }

int config_test_dir_write(struct filter_settings *current_settings, char *target_dir, char *option_name, char *start_message, char *success_message, char *failure_message_create, char *failure_message_delete)
  {
  int return_value;
  char tmp_name[MAX_BUF + 1];
  FILE *tmp_file;

  return_value = 1;

  if (target_dir != NULL)
    {
    spamdyke_log(current_settings, LOG_LEVEL_DEBUG, start_message, option_name, target_dir);

    snprintf(tmp_name, MAX_BUF, "%s/spamdyke-test_%u_%d", target_dir, (unsigned int)time(NULL), getpid());
    if ((tmp_file = fopen(tmp_name, "w")) != NULL)
      {
      fclose(tmp_file);

      if (unlink(tmp_name) == 0)
        spamdyke_log(current_settings, LOG_LEVEL_DEBUG, success_message, option_name, tmp_name);
      else
        {
        spamdyke_log(current_settings, LOG_LEVEL_ERROR, failure_message_delete, option_name, tmp_name, strerror(errno));
        return_value = 0;
        }
      }
    else
      {
      spamdyke_log(current_settings, LOG_LEVEL_ERROR, failure_message_create, option_name, tmp_name, strerror(errno));
      return_value = 0;
      }
    }

  return(return_value);
  }

int config_test_noop(struct filter_settings *current_settings, struct spamdyke_option *target_option)
  {
  return(1);
  }

int config_test_graylist(struct filter_settings *current_settings, struct spamdyke_option *target_option)
  {
  int return_value;
  int test_return;
  DIR *top_dir;
  DIR *domain_dir;
  DIR *user_dir;
  struct dirent *top_ent;
  struct dirent *domain_ent;
  struct dirent *user_ent;
  char top_name[MAX_BUF + 1];
  char domain_name[MAX_BUF + 1];
  char user_name[MAX_BUF + 1];
  int domain_count;

  return_value = 1;
  domain_count = 0;
  if ((current_settings->graylist_dir != NULL) &&
      (current_settings->graylist_dir[0] != '\0'))
    {
    spamdyke_log(current_settings, LOG_LEVEL_DEBUG, CONFIG_TEST_START_GRAYLIST, target_option->getopt_option.name, current_settings->graylist_dir);

    if ((top_dir = opendir(current_settings->graylist_dir)) != NULL)
      {
      while ((top_ent = readdir(top_dir)) != NULL)
        {
        snprintf(top_name, MAX_BUF, "%s/%s", current_settings->graylist_dir, top_ent->d_name);
        switch (top_ent->d_type)
          {
          case DT_DIR:
            if ((strcmp(top_ent->d_name, DIR_CURRENT) != 0) &&
                (strcmp(top_ent->d_name, DIR_PARENT) != 0))
              {
              domain_count++;
              test_return = config_test_dir_write(current_settings, top_name, (char *)target_option->getopt_option.name, NULL, NULL, CONFIG_TEST_ERROR_DIR_WRITE, CONFIG_TEST_ERROR_DIR_WRITE_DELETE);
              if (return_value)
                return_value = test_return;

              if ((domain_dir = opendir(top_name)) != NULL)
                {
                while ((domain_ent = readdir(domain_dir)) != NULL)
                  {
                  snprintf(domain_name, MAX_BUF, "%s/%s/%s", current_settings->graylist_dir, top_ent->d_name, domain_ent->d_name);
                  switch (domain_ent->d_type)
                    {
                    case DT_DIR:
                      if ((strcmp(domain_ent->d_name, DIR_CURRENT) != 0) &&
                          (strcmp(domain_ent->d_name, DIR_PARENT) != 0))
                        {
                        test_return = config_test_dir_write(current_settings, domain_name, (char *)target_option->getopt_option.name, NULL, NULL, CONFIG_TEST_ERROR_DIR_WRITE, CONFIG_TEST_ERROR_DIR_WRITE_DELETE);
                        if (return_value)
                          return_value = test_return;

                        if ((user_dir = opendir(domain_name)) != NULL)
                          {
                          while ((user_ent = readdir(user_dir)) != NULL)
                            {
                            snprintf(user_name, MAX_BUF, "%s/%s/%s/%s", current_settings->graylist_dir, top_ent->d_name, domain_ent->d_name, user_ent->d_name);
                            switch (user_ent->d_type)
                              {
                              case DT_DIR:
                                if ((strcmp(user_ent->d_name, DIR_CURRENT) != 0) &&
                                    (strcmp(user_ent->d_name, DIR_PARENT) != 0))
                                  {
                                  spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_ERROR_GRAYLIST_USER_SUBDIR, target_option->getopt_option.name, user_name);
                                  return_value = 0;
                                  }

                                break;
                              case DT_REG:
                                test_return = config_test_file_read_write(current_settings, user_name, (char *)target_option->getopt_option.name, NULL, NULL, CONFIG_TEST_ERROR_FILE_READ_WRITE);
                                if (return_value)
                                  return_value = test_return;

                                break;
                              default:
                                spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_ERROR_GRAYLIST_USER_OTHER, target_option->getopt_option.name, user_name);
                                return_value = 0;

                                break;
                              }
                            }

                          closedir(user_dir);
                          }
                        else
                          {
                          spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_ERROR_GRAYLIST_USER_DIR, target_option->getopt_option.name, domain_name, strerror(errno));
                          return_value = 0;
                          }
                        }

                      break;
                    case DT_REG:
                      spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_ERROR_GRAYLIST_DOMAIN_FILE, target_option->getopt_option.name, domain_name);
                      return_value = 0;

                      break;
                    default:
                      spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_ERROR_GRAYLIST_DOMAIN_OTHER, target_option->getopt_option.name, domain_name);
                      return_value = 0;

                      break;
                    }
                  }

                closedir(domain_dir);
                }
              else
                {
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_ERROR_GRAYLIST_DOMAIN_DIR, target_option->getopt_option.name, top_name, strerror(errno));
                return_value = 0;
                }
              }

            break;
          case DT_REG:
            spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_ERROR_GRAYLIST_TOP_FILE, target_option->getopt_option.name, top_name);
            return_value = 0;

            break;
          default:
            spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_ERROR_GRAYLIST_TOP_OTHER, target_option->getopt_option.name, top_name);
            return_value = 0;

            break;
          }
        }

      closedir(top_dir);

      if (domain_count == 0)
        {
        spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_ERROR_GRAYLIST_TOP_EMPTY, target_option->getopt_option.name, current_settings->graylist_dir);
        return_value = 0;
        }

      if (return_value)
        spamdyke_log(current_settings, LOG_LEVEL_DEBUG, CONFIG_TEST_SUCCESS_GRAYLIST, target_option->getopt_option.name, current_settings->graylist_dir);
      }
    else
      {
      spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_ERROR_GRAYLIST_TOP_DIR, target_option->getopt_option.name, current_settings->graylist_dir, strerror(errno));
      return_value = 0;
      }
    }

  return(return_value);
  }

int config_test_smtpauth(struct filter_settings *current_settings, struct spamdyke_option *target_option)
  {
  int return_value;
  int test_return;
  int i;
  int j;
  struct stat tmp_stat;
  struct passwd *tmp_passwd;
  char tmp_name[MAX_BUF + 1];

  return_value = 1;

  if (current_settings->smtp_auth_command != NULL)
    for (i = 0; current_settings->smtp_auth_command[i] != NULL; i++)
      {
      spamdyke_log(current_settings, LOG_LEVEL_DEBUG, CONFIG_TEST_START_SMTPAUTH, target_option->getopt_option.name, current_settings->smtp_auth_command[i]);

      for (j = 0; (current_settings->smtp_auth_command[i][j] != '\0') && !isspace((int)current_settings->smtp_auth_command[i][j]); j++);
      snprintf(tmp_name, MAX_BUF, "%.*s", j, current_settings->smtp_auth_command[i]);

      test_return = config_test_file_execute(current_settings, tmp_name, (char *)target_option->getopt_option.name, NULL, CONFIG_TEST_SUCCESS_EXECUTE, CONFIG_TEST_ERROR_EXECUTE);
      if (return_value)
        return_value = test_return;

      if (stat(tmp_name, &tmp_stat) == 0)
        {
        if (tmp_stat.st_uid != 0)
          {
          tmp_passwd = getpwuid(tmp_stat.st_uid);
          spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_WARN_SMTPAUTH_OWNER, target_option->getopt_option.name, tmp_name, (tmp_passwd != NULL) ? tmp_passwd->pw_name : LOG_MISSING_DATA, tmp_stat.st_uid);
          }
        else if ((tmp_stat.st_mode & S_ISUID) == 0)
          spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_WARN_SMTPAUTH_SETUID, target_option->getopt_option.name, tmp_name);
        }
      else
        spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_ERROR_FILE_READ, target_option->getopt_option.name, tmp_name, strerror(errno));
      }

  return(return_value);
  }

int config_test_tls_certificate(struct filter_settings *current_settings, struct spamdyke_option *target_option)
  {
  int return_value;
  int test_result;

  return_value = 1;

#ifdef HAVE_LIBSSL

  if (current_settings->tls_certificate_file != NULL)
    {
    spamdyke_log(current_settings, LOG_LEVEL_DEBUG, CONFIG_TEST_START_TLS, target_option->getopt_option.name);

    test_result = config_test_file_read(current_settings, current_settings->tls_certificate_file, (char *)target_option->getopt_option.name, NULL, CONFIG_TEST_SUCCESS_FILE_READ, CONFIG_TEST_ERROR_FILE_READ);
    if (return_value)
      return_value = test_result;

    if ((test_result = tls_test(current_settings)) == 1)
      spamdyke_log(current_settings, LOG_LEVEL_DEBUG, CONFIG_TEST_SUCCESS_TLS, target_option->getopt_option.name);

    if (return_value)
      return_value = test_result;
    }

#else /* HAVE_LIBSSL */

  test_result = 0;

  if (current_settings->tls_certificate_file != NULL)
    {
    spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_ERROR_TLS_CERT_DISABLED, target_option->getopt_option.name, current_settings->tls_certificate_file);
    return_value = 0;
    }

#endif /* HAVE_LIBSSL */

  return(return_value);
  }

int config_test_tls_privatekey(struct filter_settings *current_settings, struct spamdyke_option *target_option)
  {
  int return_value;
  int test_result;

  return_value = 1;

#ifdef HAVE_LIBSSL

  if (current_settings->tls_privatekey_file != NULL)
    {
    test_result = config_test_file_read(current_settings, current_settings->tls_privatekey_file, (char *)target_option->getopt_option.name, CONFIG_TEST_START_TLS_PRIVATEKEY, CONFIG_TEST_SUCCESS_FILE_READ, CONFIG_TEST_ERROR_FILE_READ);
    if (return_value)
      return_value = test_result;
    }

#else /* HAVE_LIBSSL */

  test_result = 0;

  if (current_settings->tls_privatekey_file != NULL)
    {
    spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_ERROR_TLS_PRIVATEKEY_DISABLED, target_option->getopt_option.name, current_settings->tls_privatekey_file);
    return_value = 0;
    }

#endif /* HAVE_LIBSSL */

  return(return_value);
  }

int config_test_tls_password(struct filter_settings *current_settings, struct spamdyke_option *target_option)
  {
  int return_value;

  return_value = 1;

#ifndef HAVE_LIBSSL

  if (current_settings->tls_privatekey_password != NULL)
    {
    spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_ERROR_TLS_PASSWORD_DISABLED, target_option->getopt_option.name);
    return_value = 0;
    }

#endif /* HAVE_LIBSSL */

  return(return_value);
  }

/*
 * Some explanation here:
 *
 * The goal is to put all of the configuration options and related data in one
 * place, instead of having default values set by one function, command line
 * options set by another function, config file options in yet another function,
 * etc.  That's the purpose of the spamdyke_option structure array.
 *
 * One of the elements of the spamdyke_option structure is a pointer to an
 * accessor function.  However, C doesn't allow anonymous functions and I don't
 * want to write a separate accessor function for every member of the
 * filter_settings structure.  So instead I decided to use a "bracketed
 * expression" like this:
 *   void (*accessor_function)(args) = ({ void tmp_func(args) { body; } &tmp_func; })
 * That works but GCC doesn't allow bracketed expressions at file scope, only
 * inside functions.
 *
 * Unfortunately, GCC also doesn't consider bracketed expressions to be
 * "constant initializers", so they can't be used to initalize a static
 * variable.
 *
 * Still more unfortunately, GCC creates nested functions (within bracketed
 * initializers) on the stack, not within the code segment.  This means pointers
 * to those functions are only valid within the scope of prepare_settings().
 * So, all of spamdyke must be run from within prepare_settings() so the options
 * array will remain valid.
 *
 * It's less obvious than just using a big switch() statement, I know.  But
 * since the "obvious" solution requires multiple switch() statements at
 * multiple places in the code and is very difficult to maintain, I believe this
 * is more elegant overall.  It will also make things much clearer as new
 * configuration options and sources are added (databases, web services,
 * Windows Registry, who knows?).
 */
int prepare_settings(int argc, char *argv[], char *envp[], int (*main_function)(struct filter_settings *, int, char **, char **))
  {
  int return_value;
  int continue_processing;
  struct filter_settings tmp_settings;
  int short_code;
  int i;
  int tmp_strlen;
  char *tmp_char;
  union
    {
    int *integer_ptr;
    char **string_ptr;
    char ***string_array_ptr;
    } ptr;
  int num_options;
  struct option *tmp_options;

  struct spamdyke_option option_list[] = {
    {
      CONFIG_TYPE_ACTION,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE,
      { "help",                               no_argument,            NULL,   'h' },
      { 0 },
      { 0 },
      { NULL },
      FILTER_DECISION_ERROR,
      FILTER_GRACE_NONE,
      NULL,
      CONFIG_ACTION(usage(current_settings, 1, NULL)),
      NULL,
      "Displays this help screen and exits"
    },
    {
      CONFIG_TYPE_ACTION,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE,
      { "version",                            no_argument,            NULL,   'v' },
      { 0 },
      { 0 },
      { NULL },
      FILTER_DECISION_ERROR,
      FILTER_GRACE_NONE,
      NULL,
      CONFIG_ACTION(usage(current_settings, 0, NULL)),
      NULL,
      "Displays the version number and copyright statement, then exits"
    },
    {
      CONFIG_TYPE_ACTION,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE,
      { "config-test",                        no_argument,            NULL,   -1  },
      { 0 },
      { 0 },
      { NULL },
      FILTER_DECISION_CONFIG_TEST,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      NULL,
      "Tests the configuration as much as possible and reports any errors to stderr that can be discovered without actually accepting an incoming message, then exits."
      " Use this option with all other options that are given during normal operation. To correctly check file permissions, run spamdyke under the account that will"
      " actually start spamdyke during processing. Default: do not test configuration"
    },
    {
      CONFIG_TYPE_ACTION,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "config-file",                        required_argument,      NULL,   'f' },
      { 0 },
      { 0 },
      { NULL },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      CONFIG_ACTION(current_return_value = process_config_file(current_settings, input_value, current_return_value, history)),
      "FILE",
      "Read additional configuration options from FILE as though they were given on the command line. Default: do not read configuration from a file"
    },
    {
      CONFIG_TYPE_BOOLEAN,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "reject-ip-in-cc-rdns",               optional_argument,      NULL,   'c' },
      { .integer_value = 0 },
      { .integer_value = 1 },
      { .get_integer = CONFIG_ACCESSOR_INTEGER(check_ip_in_rdns_cc) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      NULL,
      "Search the remote server's rDNS name for its IP address AND a two-letter country code. If both are found, reject the connection. Default: do not search"
    },
    {
      CONFIG_TYPE_BOOLEAN,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "reject-empty-rdns",                  optional_argument,      NULL,   'r' },
      { .integer_value = 0 },
      { .integer_value = 1 },
      { .get_integer = CONFIG_ACCESSOR_INTEGER(check_rdns_exist) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      NULL,
      "Reject the connection if the remote server has no rDNS name. Default: do not check for an rDNS name"
    },
    {
      CONFIG_TYPE_BOOLEAN,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "reject-unresolvable-rdns",           optional_argument,      NULL,   'R' },
      { .integer_value = 0 },
      { .integer_value = 1 },
      { .get_integer = CONFIG_ACCESSOR_INTEGER(check_rdns_resolve) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      NULL,
      "Reject the connection if the remote server's rDNS name does not resolve (search for an A record). Default: do not attempt to resolve"
    },
    {
      CONFIG_TYPE_BOOLEAN,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "reject-missing-sender-mx",           optional_argument,      NULL,   -1  },
      { .integer_value = 0 },
      { .integer_value = 1 },
      { .get_integer = CONFIG_ACCESSOR_INTEGER(check_sender_mx) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      NULL,
      "Check the domain name of the sender's email address for a mail exchanger (an MX or an A record). If neither are found, reject the connection."
      " Requires \"--local-domains-file\". Default: do not check the sender's mail exchanger"
    },
    {
      CONFIG_TYPE_INTEGER,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "log-level",                          optional_argument,      NULL,   'l' },
      { .integer_value = LOG_LEVEL_ERROR },
      { .integer_value = LOG_LEVEL_INFO },
      { .get_integer = CONFIG_ACCESSOR_INTEGER(log_level) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "LEVEL",
      "Sets the log level to LEVEL: 0 = none, 1 = errors only, 2 = errors and info, 3 = errors, info and debug messages. Default when option is not given: " STRINGIFY(LOG_LEVEL_ERROR) ","
      " Default when LEVEL is not given: " STRINGIFY(LOG_LEVEL_INFO)
    },
    {
      CONFIG_TYPE_DIR_SINGLETON,
      CONFIG_ACCESS_WRITE_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "full-log-dir",                       required_argument,      NULL,   'L' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string = CONFIG_ACCESSOR_STRING(log_dir) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "DIR",
      "Log all SMTP traffic to files in DIR. Handy for troubleshooting delivery problems but not meant to be used long-term. This option imposes a performance penalty!"
      " Default: do not log all traffic"
    },
    {
      CONFIG_TYPE_STRING_SINGLETON,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "policy-url",                         required_argument,      NULL,   'u' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string = CONFIG_ACCESSOR_STRING(policy_location) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      CONFIG_ACTION(current_settings->strlen_policy_location = (current_settings->policy_location != NULL) ? strlen(current_settings->policy_location) : 0),
      "URL",
      "Append URL to the rejection message to explain why the rejection occurred. NOTE: most servers hide rejection messages from their users and most users don't read bounce messages."
      " Maximum " STRINGIFY(MAX_POLICY) " characters. Default: no policy URL"
    },
    {
      CONFIG_TYPE_INTEGER,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "connection-timeout-secs",            required_argument,      NULL,   't' },
      { .integer_value = DEFAULT_TIMEOUT_CONNECTION_SECS },
      { .integer_value = DEFAULT_TIMEOUT_CONNECTION_SECS },
      { .get_integer = CONFIG_ACCESSOR_INTEGER(timeout_connection) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "SECS",
      "Forcibly disconnect after a total of SECS seconds, regardless of activity. A value of 0 disables this feature. Default: " STRINGIFY(DEFAULT_TIMEOUT_CONNECTION_SECS)
    },
    {
      CONFIG_TYPE_INTEGER,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "idle-timeout-secs",                  required_argument,      NULL,   'T' },
      { .integer_value = DEFAULT_TIMEOUT_COMMAND_SECS },
      { .integer_value = DEFAULT_TIMEOUT_COMMAND_SECS },
      { .get_integer = CONFIG_ACCESSOR_INTEGER(timeout_connection) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "SECS",
      "Forcibly disconnect after SECS seconds of inactivity. A value of 0 disables this feature. Default: " STRINGIFY(DEFAULT_TIMEOUT_COMMAND_SECS)
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "access-file",                        required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(access_list_file) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "FILE",
      "Use FILE to determine if the remote host is allowed to connect and/or relay. Only really needed when using spamdyke to provide SMTP AUTH for an unpatched qmail installation."
      " Most often, FILE is /etc/tcp.smtp. Default: do not use an access file"
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "local-domains-file",                 required_argument,      NULL,   'd' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(local_domains_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      CONFIG_ACTION(current_settings->local_sender = 0),
      "FILE",
      "Use FILE as a list of locally hosted domains (to determine if an email address is local or remote). Most often, FILE is /var/qmail/control/rcpthosts. Default: no domains are local"
    },
    {
      CONFIG_TYPE_INTEGER,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "max-recipients",                     required_argument,      NULL,   'a' },
      { .integer_value = 0 },
      { .integer_value = 0 },
      { .get_integer = CONFIG_ACCESSOR_INTEGER(max_rcpt_to) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "NUM",
      "Allow a maximum of NUM recipients per connection for non-local senders. Default: unlimited recipients per connection"
    },
    {
      CONFIG_TYPE_DIR_SINGLETON,
      CONFIG_ACCESS_READ_WRITE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "graylist-dir",                       required_argument,      NULL,   'g' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string = CONFIG_ACCESSOR_STRING(graylist_dir) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      &config_test_graylist,
      CONFIG_ACTION(current_settings->graylist_type = GRAYLIST_FILTER_ALL),
      "DIR",
      "Graylist all connections except those excluded by the \"never-graylist-*\" options and use DIR as a base for graylist files. DIR must contain a folder named for each domain"
      " that is to be graylisted (e.g. DIR/example.com). This option is mutually exclusive with \"no-graylist-dir\". Default: no graylisting"
    },
    {
      CONFIG_TYPE_DIR_SINGLETON,
      CONFIG_ACCESS_READ_WRITE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "no-graylist-dir",                    required_argument,      NULL,   'G' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string = CONFIG_ACCESSOR_STRING(graylist_dir) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      &config_test_noop,
      CONFIG_ACTION(current_settings->graylist_type = GRAYLIST_FILTER_NONE),
      "DIR",
      "Graylist no connections except those included by the \"always-graylist-*\" options and use DIR as a base for graylist files. DIR must contain a folder named for each domain"
      " that is to be graylisted (e.g. DIR/example.com). This option is mutually exclusive with \"graylist-dir\". Default: no graylisting"
    },
    {
      CONFIG_TYPE_INTEGER,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "graylist-min-secs",                  required_argument,      NULL,   'm' },
      { .integer_value = 0 },
      { .integer_value = 0 },
      { .get_integer = CONFIG_ACCESSOR_INTEGER(graylist_min_secs) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "SECS",
      "Require a graylist entry to be present for SECS seconds before allowing incoming mail. Requires \"graylist-dir\" or \"no-graylist-dir\". Default: no minimum time required,"
      " mail is allowed after the initial graylisting"
    },
    {
      CONFIG_TYPE_INTEGER,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "graylist-max-secs",                  required_argument,      NULL,   'M' },
      { .integer_value = 0 },
      { .integer_value = 0 },
      { .get_integer = CONFIG_ACCESSOR_INTEGER(graylist_max_secs) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "SECS",
      "Invalidate graylist entries after they are SECS seconds old. Requires \"graylist-dir\" or \"no-graylist-dir\". Default: graylist entries never expire"
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "never-graylist-ip-file",             required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(graylist_never_ip_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "FILE",
      "Never graylist connections from remote servers whose IP addresses match entries in FILE. This option only has effect when \"graylist-dir\" is given, it is ignored when"
      " \"no-graylist-dir\" is given. Default: do not exclude any connections from graylisting"
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "never-graylist-rdns-file",           required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(graylist_never_rdns_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "FILE",
      "Never graylist connections from remote servers whose rDNS names match entries in FILE. This option only has effect when \"graylist-dir\" is given, it is ignored when"
      " \"no-graylist-dir\" is given. Default: do not exclude any connections from graylisting"
    },
    {
      CONFIG_TYPE_DIR_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "never-graylist-rdns-dir",            required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(graylist_never_rdns_dir) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "DIR",
      "Never graylist connections from remote servers whose rDNS names match files in DIR. This option provides much better performance than \"never-graylist-rdns-file\" for large numbers of"
      " entries. This option only has effect when \"graylist-dir\" is given, it is ignored when \"no-graylist-dir\" is given. Default: do not exclude any connections from graylisting"
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "always-graylist-ip-file",            required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(graylist_always_ip_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "FILE",
      "Always graylist connections from remote servers whose IP addresses match entries in FILE. This option only has effect when \"no-graylist-dir\" is given, it is ignored when"
      " \"graylist-dir\" is given. Default: do not graylist any connections"
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "always-graylist-rdns-file",          required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(graylist_always_rdns_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "FILE",
      "Always graylist connections from remote servers whose rDNS names match entries in FILE. This option only has effect when \"no-graylist-dir\" is given, it is ignored when"
      " \"graylist-dir\" is given. Default: do not graylist any connections"
    },
    {
      CONFIG_TYPE_DIR_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "always-graylist-rdns-dir",           required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(graylist_always_rdns_dir) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "FILE",
      "Always graylist connections from remote servers whose rDNS names IP addresses match files in DIR. This option provides much better performance than \"always-graylist-rdns-file\" for"
      " large numbers of entries. This option only has effect when \"no-graylist-dir\" is given, it is ignored when \"graylist-dir\" is given. Default: do not graylist any connections"
    },
    {
      CONFIG_TYPE_DIR_SINGLETON,
      CONFIG_ACCESS_READ_WRITE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "greylist-dir",                       required_argument,      NULL,   'g' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string = CONFIG_ACCESSOR_STRING(graylist_dir) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      &config_test_noop,
      CONFIG_ACTION(current_settings->graylist_type = GRAYLIST_FILTER_ALL),
      NULL,
      NULL
    },
    {
      CONFIG_TYPE_DIR_SINGLETON,
      CONFIG_ACCESS_READ_WRITE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "no-greylist-dir",                    required_argument,      NULL,   'G' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string = CONFIG_ACCESSOR_STRING(graylist_dir) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      &config_test_noop,
      CONFIG_ACTION(current_settings->graylist_type = GRAYLIST_FILTER_NONE),
      NULL,
      NULL
    },
    {
      CONFIG_TYPE_INTEGER,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "greylist-min-secs",                  required_argument,      NULL,   'm' },
      { .integer_value = 0 },
      { .integer_value = 0 },
      { .get_integer = CONFIG_ACCESSOR_INTEGER(graylist_min_secs) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      NULL,
      NULL
    },
    {
      CONFIG_TYPE_INTEGER,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "greylist-max-secs",                  required_argument,      NULL,   'M' },
      { .integer_value = 0 },
      { .integer_value = 0 },
      { .get_integer = CONFIG_ACCESSOR_INTEGER(graylist_max_secs) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      NULL,
      NULL
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "never-greylist-ip-file",             required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(graylist_never_ip_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      NULL,
      NULL
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "never-greylist-rdns-file",           required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(graylist_never_rdns_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      NULL,
      NULL
    },
    {
      CONFIG_TYPE_DIR_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "never-greylist-rdns-dir",            required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(graylist_never_rdns_dir) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      NULL,
      NULL
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "always-greylist-ip-file",            required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(graylist_always_ip_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      NULL,
      NULL
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "always-greylist-rdns-file",          required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(graylist_always_rdns_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      NULL,
      NULL
    },
    {
      CONFIG_TYPE_DIR_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "always-greylist-rdns-dir",           required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(graylist_always_rdns_dir) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      NULL,
      NULL
    },
    {
      CONFIG_TYPE_COMMAND_ARRAY,
      CONFIG_ACCESS_EXECUTE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "smtp-auth-command",                  required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(smtp_auth_command) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_AFTER_FROM,
      &config_test_smtpauth,
      NULL,
      "COMMAND",
      "Perform SMTP AUTH verification using COMMAND. spamdyke will only advertise cleartext authentication methods (unless qmail has been patched to advertise encrypted methods)."
      " If the authentication is valid, all filters will be bypassed. Most often, COMMAND is \"/bin/checkpassword /bin/true\". Default: do not check authentication, even if qmail supports it"
    },
    {
      CONFIG_TYPE_COMMAND_ARRAY,
      CONFIG_ACCESS_EXECUTE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "smtp-auth-command-encryption",       required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(smtp_auth_command) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_AFTER_FROM,
      &config_test_noop,
      CONFIG_ACTION(current_settings->smtp_auth_encryption = 1),
      "COMMAND",
      "Perform SMTP AUTH verification using COMMAND. spamdyke will advertise all available authentication methods, including encrypted ones -- COMMAND must provide support for encrypted"
      " authentication. If the authentication is valid, all filters will be bypassed. Should not be used with \"smtp-auth-command\" (such usage is valid but makes no sense). Most often,"
      " COMMAND is \"/home/vpopmail/bin/vchkpw /bin/true\". Default: do not check authentication, even if qmail supports it"
    },
    {
      CONFIG_TYPE_INTEGER,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "greeting-delay-secs",                required_argument,      NULL,   'e' },
      { .integer_value = 0 },
      { .integer_value = 0 },
      { .get_integer = CONFIG_ACCESSOR_INTEGER(check_earlytalker) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "SECS",
      "Delay sending the SMTP greeting banner SECS seconds to see if the remote server begins sending data early. If it does, the connection is rejected. Default: no delay"
    },
    {
      CONFIG_TYPE_STRING_ARRAY,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "check-dnsrbl",                       required_argument,      NULL,   'x' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(dnsrbl_fqdn) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "DNSRBL",
      "Check the remote server's IP address against the realtime blackhole list DNSRBL. If it is found, the connection is rejected. Default: do not check any DNS RBLs."
    },
    {
      CONFIG_TYPE_STRING_ARRAY,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "check-dns-whitelist",                required_argument,      NULL,   -1 },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(dnsrwl_fqdn) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "DNSWHITELIST",
      "Check the remote server's IP address against the DNS whitelist DNSWHITELIST (essentially a DNSRBL that contains whitelisted IPs). If it is found, all filters are bypassed."
      " Default: do not check any DNS whitelists."
    },
    {
      CONFIG_TYPE_STRING_ARRAY,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "check-rhsbl",                        required_argument,      NULL,   'X' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(rhsbl_fqdn) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "RHSBL",
      "Check the remote server's domain name and the sender email address' domain name against the righthand-side blackhole list RHSBL. If it is found, the connection is rejected."
      " Default: do not check any DNS RBLs."
    },
    {
      CONFIG_TYPE_STRING_ARRAY,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "check-rhs-whitelist",                required_argument,      NULL,   -1 },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(rhswl_fqdn) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_AFTER_FROM,
      NULL,
      NULL,
      "RHSWHITELIST",
      "Check the remote server's domain name and the sender email address' domain name against the righthand-side whitelist RHSWHITELIST (essentially an RHSBL that contains whitelisted"
      " domains). If it is found, all filters are bypassed. Default: do not check any RHS whitelists."
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "ip-in-rdns-keyword-file",            required_argument,      NULL,   'k' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(blacklist_rdns_keyword_file) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "FILE",
      "Search the remote server's rDNS name for its IP address AND a keyword in FILE. If both are found, the connection is rejected. Default: do not search"
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "rdns-blacklist-file",                required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(blacklist_rdns_file) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "FILE",
      "Reject the connection if the remote server's rDNS name matches an entry in FILE. Default: do not search"
    },
    {
      CONFIG_TYPE_DIR_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "rdns-blacklist-dir",                 required_argument,      NULL,   'b' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(blacklist_rdns_dir) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "DIR",
      "Reject the connection if the remote server's rDNS name matches a file in DIR. This option provides much better performance than \"blacklist_rdns_file\" for large numbers of entries."
      " Default: do not search"
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "ip-blacklist-file",                  required_argument,      NULL,   'B' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(blacklist_ip_file) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "FILE",
      "Reject the connection if the remote server's IP address matches an entry in FILE. Default: do not search"
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "rdns-whitelist-file",                required_argument,      NULL,   'w' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(whitelist_rdns_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "FILE",
      "If the remote server's rDNS name matches an entry in FILE, bypass all filters. Default: do not search"
    },
    {
      CONFIG_TYPE_DIR_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "rdns-whitelist-dir",                 required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(whitelist_rdns_dir) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "DIR",
      "If the remote server's rDNS name matches a file in DIR, bypass all filters. Default: do not search"
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "ip-whitelist-file",                  required_argument,      NULL,   'W' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(whitelist_ip_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "FILE",
      "If the remote server's IP address matches an entry in FILE, bypass all filters. Default: do not search"
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "sender-blacklist-file",              required_argument,      NULL,   's' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(blacklist_sender_file) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "FILE",
      "Reject the connection if the sender's email address matches an entry in FILE. Default: do not search"
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "sender-whitelist-file",              required_argument,      NULL,   -1 },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(whitelist_sender_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_AFTER_FROM,
      NULL,
      NULL,
      "FILE",
      "If the sender's email address matches an entry in FILE, bypass all filters. Default: do not search"
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "recipient-blacklist-file",           required_argument,      NULL,   'S' },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(blacklist_recipient_file) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "FILE",
      "Reject any recipient addresses that match entries in FILE. Default: do not search"
    },
    {
      CONFIG_TYPE_FILE_ARRAY,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "recipient-whitelist-file",           required_argument,      NULL,   -1 },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string_array = CONFIG_ACCESSOR_STRING_ARRAY(whitelist_recipient_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_AFTER_TO,
      NULL,
      NULL,
      "FILE",
      "If the recipient's email address matches an entry in FILE, bypass all filters. Default: do not search"
  },
    {
      CONFIG_TYPE_STRING_SINGLETON,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "hostname",                           required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string = CONFIG_ACCESSOR_STRING(local_server_name) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "NAME",
      "Use NAME as the fully qualified domain name of this host. This value is only used to create an encrypted challenge during SMTP AUTH challenge-response."
      " Default: " MISSING_LOCAL_SERVER_NAME
    },
    {
      CONFIG_TYPE_FILE_SINGLETON,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "hostname-file",                      required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string = CONFIG_ACCESSOR_STRING(local_server_name_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "FILE",
      "Read the fully qualified domain name of this host from the first line of FILE. This value is only used to create an encrypted challenge during SMTP AUTH challenge-response."
      " This option is ignored if \"hostname\" is given. Default: do not read the hostname from a file"
    },
    {
      CONFIG_TYPE_COMMAND_SINGLETON,
      CONFIG_ACCESS_EXECUTE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "hostname-command",                   required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string = CONFIG_ACCESSOR_STRING(local_server_name_command) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      NULL,
      NULL,
      "COMMAND",
      "Read the fully qualified domain name of this host from the output of COMMAND. Most often, this value is \"/bin/hostname -f\". This value is only used to create an encrypted"
      " challenge during SMTP AUTH challenge-response. This option is ignored if \"hostname\" or \"hostname-file\" are given. Default: do not read the hostname from a command"
    },
    {
      CONFIG_TYPE_FILE_SINGLETON,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "tls-certificate-file",               required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string = CONFIG_ACCESSOR_STRING(tls_certificate_file) },
      FILTER_DECISION_DO_FILTER,
      FILTER_GRACE_NONE,
      &config_test_tls_certificate,
      NULL,
      "FILE",
      "Offer TLS support using the SSL certificate in FILE. FILE must be in PEM format. If FILE does not also contain the private key, \"tls-privatekey-file\" must be used."
      " Default: do not support TLS; if qmail supports TLS, simply pass the encrypted traffic without interference (some filters will be bypassed by this action)."
    },
    {
      CONFIG_TYPE_FILE_SINGLETON,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "tls-privatekey-file",                required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string = CONFIG_ACCESSOR_STRING(tls_privatekey_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      &config_test_tls_privatekey,
      NULL,
      "FILE",
      "Read the private key for the SSL certificate (from \"tls-certificate-file\") from FILE. FILE must be in PEM format. Requires \"tls-certificate-file\"."
      " Default: find the private key in the certificate file given with \"tls-certificate-file\""
    },
    {
      CONFIG_TYPE_STRING_SINGLETON,
      CONFIG_ACCESS_NONE,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "tls-privatekey-password",            required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string = CONFIG_ACCESSOR_STRING(tls_privatekey_password) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      &config_test_tls_password,
      CONFIG_ACTION(current_settings->tls_strlen_privatekey_password = (current_settings->tls_privatekey_password != NULL) ? strlen(current_settings->tls_privatekey_password) : 0),
      "PASSWORD",
      "Use PASSWORD to decrypt the SSL private key (from \"tls-certificate-file\" or \"tls-privatekey-file\"), if necessary. NOTE: this option reveals the password in the process list!"
      " Requires \"tls-certificate-file\" and/or \"tls-privatekey-file\". Default: assume the private key is not encrypted with a password."
    },
    {
      CONFIG_TYPE_FILE_SINGLETON,
      CONFIG_ACCESS_READ_ONLY,
      CONFIG_LOCATION_CMDLINE | CONFIG_LOCATION_FILE,
      { "tls-privatekey-password-file",       required_argument,      NULL,   -1  },
      { .string_value = NULL },
      { .string_value = NULL },
      { .get_string = CONFIG_ACCESSOR_STRING(tls_privatekey_password_file) },
      FILTER_DECISION_UNDECIDED,
      FILTER_GRACE_NONE,
      &config_test_tls_password,
      NULL,
      "FILE",
      "Read the password to decrypt the private key for the SSL certificate (from \"tls-certificate-file\") from the first line of FILE, if necessary. Requires \"tls-certificate-file\""
      " and/or \"tls-password-file\". Default: assume the private key is not encrypted with a password."
    },

    { CONFIG_TYPE_NONE }
    };

  short_code = CONFIG_SHORT_CODE_START;
  for (i = 0; option_list[i].value_type != CONFIG_TYPE_NONE; i++)
    if (option_list[i].getopt_option.val == -1)
      {
      option_list[i].getopt_option.val = short_code;
      short_code++;
      }

  continue_processing = 1;

  tmp_settings.option_list = option_list;

  tmp_settings.filter_action = FILTER_DECISION_UNDECIDED;
  tmp_settings.filter_grace = FILTER_GRACE_AFTER_FROM;
  tmp_settings.log_target = LOG_USE_SYSLOG;

  tmp_settings.server_name[0] = '\0';
  tmp_settings.strlen_server_name = 0;
  tmp_settings.server_ip = NULL;
  tmp_settings.strlen_server_ip = 0;
  tmp_settings.local_server_name = NULL;

  tmp_settings.smtp_auth_encryption = 0;

  tmp_settings.graylist_type = GRAYLIST_NONE;

  tmp_settings.allow_relay = 1;
  tmp_settings.inside_data = 0;
  tmp_settings.num_rcpt_to = 0;
  tmp_settings.local_sender = 1;

  tmp_settings.sender_address[0] = '\0';
  tmp_settings.recipient_address[0] = '\0';
  tmp_settings.graylist_path[0] = '\0';

  tmp_settings.rejection = NULL;
  tmp_settings.transient_rejection = NULL;
  tmp_settings.reject_message_buf[0] = '\0';
  tmp_settings.transient_reject_message_buf[0] = '\0';
  tmp_settings.short_reject_message_buf[0] = '\0';

  tmp_settings.child_argv = NULL;

  tmp_settings.smtp_auth_type = SMTP_AUTH_UNKNOWN;
  tmp_settings.smtp_auth_origin = SMTP_AUTH_ORIGIN_NONE;
  tmp_settings.smtp_auth_state = SMTP_AUTH_STATE_NONE;
  tmp_settings.smtp_auth_challenge[0] = '\0';
  tmp_settings.smtp_auth_response[0] = '\0';
  tmp_settings.smtp_auth_username[0] = '\0';

  tmp_settings.tls_state = TLS_STATE_INACTIVE;
  tmp_settings.tls_strlen_privatekey_password = 0;

#ifdef HAVE_LIBSSL

  tmp_settings.tls_context = NULL;
  tmp_settings.tls_session = NULL;

#endif /* HAVE_LIBSSL */

  for (i = 0; option_list[i].value_type != CONFIG_TYPE_NONE; i++)
    switch (option_list[i].value_type)
      {
      case CONFIG_TYPE_BOOLEAN:
      case CONFIG_TYPE_INTEGER:
        if ((option_list[i].getter.get_integer != NULL) &&
            ((ptr.integer_ptr = (*option_list[i].getter.get_integer)(&tmp_settings)) != NULL))
          *ptr.integer_ptr = option_list[i].default_value.integer_value;

        break;
      case CONFIG_TYPE_STRING_SINGLETON:
      case CONFIG_TYPE_FILE_SINGLETON:
      case CONFIG_TYPE_DIR_SINGLETON:
      case CONFIG_TYPE_COMMAND_SINGLETON:
        if ((option_list[i].getter.get_string != NULL) &&
            ((ptr.string_ptr = (*option_list[i].getter.get_string)(&tmp_settings)) != NULL))
          {
          *ptr.string_ptr = NULL;
          if (option_list[i].default_value.string_value != NULL)
            {
          tmp_strlen = strlen(option_list[i].default_value.string_value);
            if ((tmp_char = (char *)malloc(tmp_strlen + 1)) != NULL)
              {
              strncpy(tmp_char, option_list[i].default_value.string_value, tmp_strlen);
              tmp_char[tmp_strlen] = '\0';
              *ptr.string_ptr = tmp_char;
              }
            else
              {
              spamdyke_log(&tmp_settings, LOG_LEVEL_ERROR, LOG_ERROR_REALLOC, (unsigned long)(strlen(option_list[i].default_value.string_value) + 1));
              continue_processing = 0;
              }
            }
          }

        break;
      case CONFIG_TYPE_STRING_ARRAY:
      case CONFIG_TYPE_FILE_ARRAY:
      case CONFIG_TYPE_DIR_ARRAY:
      case CONFIG_TYPE_COMMAND_ARRAY:
        if ((option_list[i].getter.get_string_array != NULL) &&
            ((ptr.string_array_ptr = (*option_list[i].getter.get_string_array)(&tmp_settings)) != NULL))
          {
          *ptr.string_array_ptr = NULL;
          if (option_list[i].default_value.string_value != NULL)
            continue_processing = append_string(NULL, ptr.string_array_ptr, option_list[i].default_value.string_value, strlen(option_list[i].default_value.string_value));
          }
        break;
      }

  num_options = 0;
  for (i = 0; option_list[i].value_type != CONFIG_TYPE_NONE; i++)
    if (option_list[i].location & CONFIG_LOCATION_CMDLINE)
      num_options++;

  if ((tmp_options = (struct option *)malloc(sizeof(struct option) * (num_options + 1))) != NULL)
    {
    num_options = 0;
    for (i = 0; option_list[i].value_type != CONFIG_TYPE_NONE; i++)
      if (option_list[i].location & CONFIG_LOCATION_CMDLINE)
        {
        memcpy(&tmp_options[num_options], &option_list[i].getopt_option, sizeof(struct option));
        num_options++;
        }

    memcpy(&tmp_options[num_options], &option_list[i].getopt_option, sizeof(struct option));
    tmp_settings.long_options = tmp_options;
    }

  tmp_settings.short_options[0] = '+';
  tmp_settings.short_options[1] = '\0';
  tmp_strlen = 1;
  for (i = 0; (option_list[i].value_type != CONFIG_TYPE_NONE) && (tmp_strlen < MAX_BUF); i++)
    if ((option_list[i].location & CONFIG_LOCATION_CMDLINE) &&
        (option_list[i].getopt_option.val > 0) &&
        (option_list[i].getopt_option.val <= 255) &&
        isalnum((int)option_list[i].getopt_option.val) &&
        (strchr(tmp_settings.short_options, option_list[i].getopt_option.val) == NULL))
      {
      tmp_settings.short_options[tmp_strlen] = option_list[i].getopt_option.val;
      tmp_strlen++;

      if ((option_list[i].getopt_option.has_arg == required_argument) &&
          (tmp_strlen < MAX_BUF))
        {
        tmp_settings.short_options[tmp_strlen] = ':';
        tmp_strlen++;
        }
      else if ((option_list[i].getopt_option.has_arg == optional_argument) &&
               ((tmp_strlen + 1) < MAX_BUF))
        {
        tmp_settings.short_options[tmp_strlen] = ':';
        tmp_settings.short_options[tmp_strlen + 1] = ':';
        tmp_strlen += 2;
        }

      tmp_settings.short_options[tmp_strlen] = '\0';
      }

  if (continue_processing &&
      (main_function != NULL))
    return_value = (*main_function)(&tmp_settings, argc, argv, envp);
  else
    return_value = 111;

  return(return_value);
  }

int append_string(struct filter_settings *current_settings, char ***target_list, char *target_string, int strlen_target_string)
  {
  int return_value;
  int i;
  char *tmp_char;
  char **tmp_array;

  return_value = 0;

  if ((*target_list) != NULL)
    for (i = 0; (*target_list)[i] != NULL; i++);
  else
    i = 0;

  tmp_char = NULL;
  if ((strlen_target_string == -1) ||
      ((tmp_char = malloc(strlen_target_string + 1)) != NULL))
    if ((tmp_array = realloc((*target_list), sizeof(char *) * (i + 2))) != NULL)
      {
      if (strlen_target_string >= 0)
        {
        memcpy(tmp_char, target_string, strlen_target_string);
        tmp_char[strlen_target_string] = '\0';
        tmp_array[i] = tmp_char;
        }
      else
        tmp_array[i] = target_string;

      tmp_array[i + 1] = NULL;
      *target_list = tmp_array;

      return_value = 1;
      }
    else
      {
      spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_REALLOC, (unsigned long)(sizeof(char *) * (i + 2)));
      free(tmp_char);
      }
  else
    spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_REALLOC, (unsigned long)(strlen_target_string + 1));

  return(return_value);
  }

int test_settings(struct filter_settings *current_settings)
  {
  int return_value;
  int test_return;
  int i;
  int j;
  int k;
  union
    {
    int *integer_ptr;
    char **string_ptr;
    char ***string_array_ptr;
    } ptr;
  uid_t tmp_uid;
  char tmp_name[MAX_BUF + 1];
  struct passwd *tmp_passwd;

  usage(current_settings, 0, NULL);

  current_settings->log_level = MAXVAL(LOG_LEVEL_ERROR, current_settings->log_level);
  current_settings->log_target = LOG_USE_STDERR;
  spamdyke_log(current_settings, LOG_LEVEL_INFO, CONFIG_TEST_START);

  if ((tmp_uid = geteuid()) != 0)
    {
    tmp_passwd = getpwuid(tmp_uid);
    spamdyke_log(current_settings, LOG_LEVEL_INFO, CONFIG_TEST_SUCCESS_UID, (tmp_passwd != NULL) ? tmp_passwd->pw_name : LOG_MISSING_DATA, tmp_uid);
    }
  else
    spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_ERROR_UID);

  return_value = 1;
  for (i = 0; current_settings->option_list[i].value_type != CONFIG_TYPE_NONE; i++)
    if (current_settings->option_list[i].test_function != NULL)
      return_value = (!(*current_settings->option_list[i].test_function)(current_settings, current_settings->option_list + i)) ? 0 : return_value;
    else
      switch (current_settings->option_list[i].value_type)
        {
        case CONFIG_TYPE_FILE_SINGLETON:
          if ((current_settings->option_list[i].getter.get_string != NULL) &&
              ((ptr.string_ptr = (*current_settings->option_list[i].getter.get_string)(current_settings)) != NULL) &&
              (*ptr.string_ptr != NULL))
            switch (current_settings->option_list[i].access_type)
              {
              case CONFIG_ACCESS_READ_ONLY:
                test_return = config_test_file_read(current_settings, *ptr.string_ptr, (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_FILE_READ, CONFIG_TEST_SUCCESS_FILE_READ, CONFIG_TEST_ERROR_FILE_READ);
                if (return_value)
                  return_value = test_return;

                break;
              case CONFIG_ACCESS_WRITE_ONLY:
                test_return = config_test_file_write(current_settings, *ptr.string_ptr, (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_FILE_WRITE, CONFIG_TEST_SUCCESS_FILE_WRITE, CONFIG_TEST_ERROR_FILE_WRITE);
                if (return_value)
                  return_value = test_return;

                break;
              case CONFIG_ACCESS_READ_WRITE:
                test_return = config_test_file_read_write(current_settings, *ptr.string_ptr, (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_FILE_READ_WRITE, CONFIG_TEST_SUCCESS_FILE_READ_WRITE, CONFIG_TEST_ERROR_FILE_READ_WRITE);
                if (return_value)
                  return_value = test_return;

                break;
              case CONFIG_ACCESS_EXECUTE:
                test_return = config_test_file_execute(current_settings, *ptr.string_ptr, (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_EXECUTE, CONFIG_TEST_SUCCESS_EXECUTE, CONFIG_TEST_ERROR_EXECUTE);
                if (return_value)
                  return_value = test_return;

                break;
              default:
                break;
              }

          break;
        case CONFIG_TYPE_FILE_ARRAY:
          if ((current_settings->option_list[i].getter.get_string_array != NULL) &&
              ((ptr.string_array_ptr = (*current_settings->option_list[i].getter.get_string_array)(current_settings)) != NULL) &&
              (*ptr.string_array_ptr != NULL))
            switch (current_settings->option_list[i].access_type)
              {
              case CONFIG_ACCESS_READ_ONLY:
                for (j = 0; (*ptr.string_array_ptr)[j] != NULL; j++)
                  {
                  test_return = config_test_file_read(current_settings, (*ptr.string_array_ptr)[j], (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_FILE_READ, CONFIG_TEST_SUCCESS_FILE_READ, CONFIG_TEST_ERROR_FILE_READ);
                  if (return_value)
                    return_value = test_return;
                  }

                break;
              case CONFIG_ACCESS_WRITE_ONLY:
                for (j = 0; (*ptr.string_array_ptr)[j] != NULL; j++)
                  {
                  test_return = config_test_file_write(current_settings, (*ptr.string_array_ptr)[j], (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_FILE_WRITE, CONFIG_TEST_SUCCESS_FILE_WRITE, CONFIG_TEST_ERROR_FILE_WRITE);
                  if (return_value)
                    return_value = test_return;
                  }

                break;
              case CONFIG_ACCESS_READ_WRITE:
                for (j = 0; (*ptr.string_array_ptr)[j] != NULL; j++)
                  {
                  test_return = config_test_file_read_write(current_settings, (*ptr.string_array_ptr)[j], (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_FILE_READ_WRITE, CONFIG_TEST_SUCCESS_FILE_READ_WRITE, CONFIG_TEST_ERROR_FILE_READ_WRITE);
                  if (return_value)
                    return_value = test_return;
                  }

                break;
              case CONFIG_ACCESS_EXECUTE:
                for (j = 0; (*ptr.string_array_ptr)[j] != NULL; j++)
                  {
                  test_return = config_test_file_execute(current_settings, (*ptr.string_array_ptr)[j], (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_EXECUTE, CONFIG_TEST_SUCCESS_EXECUTE, CONFIG_TEST_ERROR_EXECUTE);
                  if (return_value)
                    return_value = test_return;
                  }

                break;
              default:
                break;
              }

          break;
        case CONFIG_TYPE_DIR_SINGLETON:
          if ((current_settings->option_list[i].getter.get_string != NULL) &&
              ((ptr.string_ptr = (*current_settings->option_list[i].getter.get_string)(current_settings)) != NULL) &&
              (*ptr.string_ptr != NULL))
            switch (current_settings->option_list[i].access_type)
              {
              case CONFIG_ACCESS_READ_ONLY:
                test_return = config_test_dir_read(current_settings, *ptr.string_ptr, (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_DIR_READ, CONFIG_TEST_SUCCESS_DIR_READ, CONFIG_TEST_ERROR_DIR_READ);
                if (return_value)
                  return_value = test_return;

                break;
              case CONFIG_ACCESS_WRITE_ONLY:
                test_return = config_test_dir_write(current_settings, *ptr.string_ptr, (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_DIR_WRITE, CONFIG_TEST_SUCCESS_DIR_WRITE, CONFIG_TEST_ERROR_DIR_WRITE, CONFIG_TEST_ERROR_DIR_WRITE_DELETE);
                if (return_value)
                  return_value = test_return;

                break;
              case CONFIG_ACCESS_READ_WRITE:
                test_return = config_test_dir_read(current_settings, *ptr.string_ptr, (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_DIR_READ, CONFIG_TEST_SUCCESS_DIR_READ, CONFIG_TEST_ERROR_DIR_READ);
                if (return_value)
                  return_value = test_return;

                test_return = config_test_dir_write(current_settings, *ptr.string_ptr, (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_DIR_WRITE, CONFIG_TEST_SUCCESS_DIR_WRITE, CONFIG_TEST_ERROR_DIR_WRITE, CONFIG_TEST_ERROR_DIR_WRITE_DELETE);
                if (return_value)
                  return_value = test_return;

                break;
              case CONFIG_ACCESS_EXECUTE:
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_BAD_CONFIG_DIR_EXEC, (char *)current_settings->option_list[i].getopt_option.name);
                return_value = 0;

                break;
              default:
                break;
              }

          break;
        case CONFIG_TYPE_DIR_ARRAY:
          if ((current_settings->option_list[i].getter.get_string_array != NULL) &&
              ((ptr.string_array_ptr = (*current_settings->option_list[i].getter.get_string_array)(current_settings)) != NULL) &&
              (*ptr.string_array_ptr != NULL))
            switch (current_settings->option_list[i].access_type)
              {
              case CONFIG_ACCESS_READ_ONLY:
                for (j = 0; (*ptr.string_array_ptr)[j] != NULL; j++)
                  {
                  test_return = config_test_dir_read(current_settings, (*ptr.string_array_ptr)[j], (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_DIR_READ, CONFIG_TEST_SUCCESS_DIR_READ, CONFIG_TEST_ERROR_DIR_READ);
                  if (return_value)
                    return_value = test_return;
                  }

                break;
              case CONFIG_ACCESS_WRITE_ONLY:
                for (j = 0; (*ptr.string_array_ptr)[j] != NULL; j++)
                  {
                  test_return = config_test_dir_write(current_settings, (*ptr.string_array_ptr)[j], (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_DIR_WRITE, CONFIG_TEST_SUCCESS_DIR_WRITE, CONFIG_TEST_ERROR_DIR_WRITE, CONFIG_TEST_ERROR_DIR_WRITE_DELETE);
                  if (return_value)
                    return_value = test_return;
                  }

                break;
              case CONFIG_ACCESS_READ_WRITE:
                for (j = 0; (*ptr.string_array_ptr)[j] != NULL; j++)
                  {
                  test_return = config_test_dir_read(current_settings, (*ptr.string_array_ptr)[j], (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_DIR_READ, CONFIG_TEST_SUCCESS_DIR_READ, CONFIG_TEST_ERROR_DIR_READ);
                  if (return_value)
                    return_value = test_return;

                  test_return = config_test_dir_write(current_settings, (*ptr.string_array_ptr)[j], (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_DIR_WRITE, CONFIG_TEST_SUCCESS_DIR_WRITE, CONFIG_TEST_ERROR_DIR_WRITE, CONFIG_TEST_ERROR_DIR_WRITE_DELETE);
                  if (return_value)
                    return_value = test_return;
                  }

                break;
              case CONFIG_ACCESS_EXECUTE:
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_BAD_CONFIG_DIR_EXEC, (char *)current_settings->option_list[i].getopt_option.name);
                return_value = 0;

                break;
              default:
                break;
              }

          break;
        case CONFIG_TYPE_COMMAND_SINGLETON:
          if ((current_settings->option_list[i].getter.get_string != NULL) &&
              ((ptr.string_ptr = (*current_settings->option_list[i].getter.get_string)(current_settings)) != NULL) &&
              (*ptr.string_ptr != NULL))
            switch (current_settings->option_list[i].access_type)
              {
              case CONFIG_ACCESS_READ_ONLY:
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_BAD_CONFIG_CMD_READ, (char *)current_settings->option_list[i].getopt_option.name);
                return_value = 0;

                break;
              case CONFIG_ACCESS_WRITE_ONLY:
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_BAD_CONFIG_CMD_WRITE, (char *)current_settings->option_list[i].getopt_option.name);
                return_value = 0;

                break;
              case CONFIG_ACCESS_READ_WRITE:
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_BAD_CONFIG_CMD_READ_WRITE, (char *)current_settings->option_list[i].getopt_option.name);
                return_value = 0;

                break;
              case CONFIG_ACCESS_EXECUTE:
                for (k = 0; ((*ptr.string_ptr)[k] != '\0') && !isspace((int)(*ptr.string_ptr)[k]); k++);
                snprintf(tmp_name, MAX_BUF, "%.*s", k, *ptr.string_ptr);

                test_return = config_test_file_execute(current_settings, tmp_name, (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_EXECUTE, CONFIG_TEST_SUCCESS_EXECUTE, CONFIG_TEST_ERROR_EXECUTE);
                if (return_value)
                  return_value = test_return;

                break;
              default:
                break;
              }

          break;
        case CONFIG_TYPE_COMMAND_ARRAY:
          if ((current_settings->option_list[i].getter.get_string_array != NULL) &&
              ((ptr.string_array_ptr = (*current_settings->option_list[i].getter.get_string_array)(current_settings)) != NULL) &&
              (*ptr.string_array_ptr != NULL))
            switch (current_settings->option_list[i].access_type)
              {
              case CONFIG_ACCESS_READ_ONLY:
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_BAD_CONFIG_CMD_READ, (char *)current_settings->option_list[i].getopt_option.name);
                return_value = 0;

                break;
              case CONFIG_ACCESS_WRITE_ONLY:
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_BAD_CONFIG_CMD_WRITE, (char *)current_settings->option_list[i].getopt_option.name);
                return_value = 0;

                break;
              case CONFIG_ACCESS_READ_WRITE:
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, CONFIG_TEST_BAD_CONFIG_CMD_READ_WRITE, (char *)current_settings->option_list[i].getopt_option.name);
                return_value = 0;

                break;
              case CONFIG_ACCESS_EXECUTE:
                for (j = 0; (*ptr.string_array_ptr)[j] != NULL; j++)
                  {
                  for (k = 0; ((*ptr.string_array_ptr)[j][k] != '\0') && !isspace((int)(*ptr.string_array_ptr)[j][k]); k++);
                  snprintf(tmp_name, MAX_BUF, "%.*s", k, (*ptr.string_array_ptr)[j]);

                  test_return = config_test_file_execute(current_settings, tmp_name, (char *)current_settings->option_list[i].getopt_option.name, CONFIG_TEST_START_EXECUTE, CONFIG_TEST_SUCCESS_EXECUTE, CONFIG_TEST_ERROR_EXECUTE);
                  if (return_value)
                    return_value = test_return;
                  }

                break;
              default:
                break;
              }

          break;
        default:
          break;
        }

  spamdyke_log(current_settings, LOG_LEVEL_ERROR, return_value ? CONFIG_TEST_SUCCESS : CONFIG_TEST_ERROR);

  return(return_value);
  }

int set_config_value(struct filter_settings *current_settings, struct spamdyke_option *target_option, char *input_value, int current_return_value, struct previous_action *history)
  {
  int return_value;
  int i;
  int j;
  int tmp_int;
  int changed_value;
  int found_match;
  char *tmp_char;
  char *tmp_input_value;
  union
    {
    int *integer_ptr;
    char **string_ptr;
    char ***string_array_ptr;
    } ptr;

  return_value = current_return_value;
  changed_value = 0;

  if (target_option != NULL)
    switch (target_option->value_type)
      {
      case CONFIG_TYPE_ACTION:
        changed_value = 1;

        if (current_settings->filter_grace < target_option->set_grace)
          current_settings->filter_grace = target_option->set_grace;

        if (return_value < target_option->set_consequence)
          return_value = target_option->set_consequence;

        break;
      case CONFIG_TYPE_BOOLEAN:
        if ((target_option->getter.get_integer != NULL) &&
            ((ptr.integer_ptr = (*target_option->getter.get_integer)(current_settings)) != NULL))
          if (input_value != NULL)
            {
            if (strchr(COMMAND_LINE_TRUE, input_value[0]) != NULL)
              {
              if ((*ptr.integer_ptr) != 1)
                {
                *ptr.integer_ptr = 1;
                changed_value = 1;

                if (current_settings->filter_grace < target_option->set_grace)
                  current_settings->filter_grace = target_option->set_grace;

                if (return_value < target_option->set_consequence)
                  return_value = target_option->set_consequence;
                }
              }
            else if (strchr(COMMAND_LINE_FALSE, input_value[0]) != NULL)
              {
              if ((*ptr.integer_ptr) != 0)
                {
                *ptr.integer_ptr = 0;
                changed_value = 1;

                if (current_settings->filter_grace < target_option->set_grace)
                  current_settings->filter_grace = target_option->set_grace;

                if (return_value < target_option->set_consequence)
                  return_value = target_option->set_consequence;
                }
              }
            else
              {
              usage(current_settings, 0, ERROR_CONFIG_BAD_VALUE, target_option->getopt_option.name, input_value);
              return_value = FILTER_DECISION_ERROR;
              }
            }
          else if ((*ptr.integer_ptr) != target_option->missing_value.integer_value)
            {
            *ptr.integer_ptr = target_option->missing_value.integer_value;
            changed_value = 1;

            if (current_settings->filter_grace < target_option->set_grace)
              current_settings->filter_grace = target_option->set_grace;

            if (return_value < target_option->set_consequence)
              return_value = target_option->set_consequence;
            }
          else
            {
            usage(current_settings, 0, ERROR_CONFIG_BAD_VALUE, target_option->getopt_option.name, input_value);
            return_value = FILTER_DECISION_ERROR;
            }
        else
          {
          usage(current_settings, 0, ERROR_CONFIG_UNKNOWN_OPTION, target_option->getopt_option.name);
          return_value = FILTER_DECISION_ERROR;
          }

        break;
      case CONFIG_TYPE_INTEGER:
        if ((target_option->getter.get_integer != NULL) &&
            ((ptr.integer_ptr = (*target_option->getter.get_integer)(current_settings)) != NULL))
          if (input_value != NULL)
            if (sscanf(input_value, "%d", &tmp_int) == 1)
              {
              if ((*ptr.integer_ptr) != tmp_int)
                {
                *ptr.integer_ptr = tmp_int;
                changed_value = 1;

                if (current_settings->filter_grace < target_option->set_grace)
                  current_settings->filter_grace = target_option->set_grace;

                if (return_value < target_option->set_consequence)
                  return_value = target_option->set_consequence;
                }
              }
            else
              {
              usage(current_settings, 0, ERROR_CONFIG_BAD_VALUE, target_option->getopt_option.name, input_value);
              return_value = FILTER_DECISION_ERROR;
              }
          else if ((*ptr.integer_ptr) != target_option->missing_value.integer_value)
            {
            *ptr.integer_ptr = target_option->missing_value.integer_value;
            changed_value = 1;

            if (current_settings->filter_grace < target_option->set_grace)
              current_settings->filter_grace = target_option->set_grace;

            if (return_value < target_option->set_consequence)
              return_value = target_option->set_consequence;
            }
          else
            {
            usage(current_settings, 0, ERROR_CONFIG_BAD_VALUE, target_option->getopt_option.name, input_value);
            return_value = FILTER_DECISION_ERROR;
            }
        else
          {
          usage(current_settings, 0, ERROR_CONFIG_UNKNOWN_OPTION, target_option->getopt_option.name);
          return_value = FILTER_DECISION_ERROR;
          }

        break;
      case CONFIG_TYPE_STRING_SINGLETON:
      case CONFIG_TYPE_FILE_SINGLETON:
      case CONFIG_TYPE_DIR_SINGLETON:
      case CONFIG_TYPE_COMMAND_SINGLETON:
        if ((target_option->getter.get_string != NULL) &&
            ((ptr.string_ptr = (*target_option->getter.get_string)(current_settings)) != NULL))
          {
          tmp_input_value = (input_value != NULL) ? input_value : target_option->missing_value.string_value;
          if (tmp_input_value != NULL)
            {
            if (((*ptr.string_ptr) == NULL) ||
                (strcmp(*ptr.string_ptr, tmp_input_value) != 0))
              {
              tmp_int = strlen(tmp_input_value);
              if ((tmp_char = (char *)realloc(*ptr.string_ptr, tmp_int + 1)) != NULL)
                {
                strncpy(tmp_char, tmp_input_value, tmp_int);
                tmp_char[tmp_int] = '\0';
                *ptr.string_ptr = tmp_char;
                changed_value = 1;

                if (current_settings->filter_grace < target_option->set_grace)
                  current_settings->filter_grace = target_option->set_grace;

                if (return_value < target_option->set_consequence)
                  return_value = target_option->set_consequence;
                }
              else
                {
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_REALLOC, (unsigned long)(strlen(tmp_input_value) + 1));
                return_value = FILTER_DECISION_ERROR;
                }
              }
            }
          else if ((*ptr.string_ptr) != NULL)
            {
            free(*ptr.string_ptr);
            *ptr.string_ptr = NULL;
            changed_value = 1;

            if (current_settings->filter_grace < target_option->set_grace)
              current_settings->filter_grace = target_option->set_grace;

            if (return_value < target_option->set_consequence)
              return_value = target_option->set_consequence;
            }
          else
            {
            usage(current_settings, 0, ERROR_CONFIG_BAD_VALUE, target_option->getopt_option.name, input_value);
            return_value = FILTER_DECISION_ERROR;
            }
          }
        else
          {
          usage(current_settings, 0, ERROR_CONFIG_UNKNOWN_OPTION, target_option->getopt_option.name);
          return_value = FILTER_DECISION_ERROR;
          }

        if (changed_value &&
            (target_option->value_type == CONFIG_TYPE_COMMAND_SINGLETON) &&
            ((*ptr.string_ptr) != NULL))
          for (i = 0; (*ptr.string_ptr)[i] != '\0'; i++)
            if ((*ptr.string_ptr)[i] == COMMAND_LINE_SPACER)
              (*ptr.string_ptr)[i] = ' ';

        break;
      case CONFIG_TYPE_STRING_ARRAY:
      case CONFIG_TYPE_FILE_ARRAY:
      case CONFIG_TYPE_DIR_ARRAY:
      case CONFIG_TYPE_COMMAND_ARRAY:
        if ((target_option->getter.get_string_array != NULL) &&
            ((ptr.string_array_ptr = (*target_option->getter.get_string_array)(current_settings)) != NULL))
          {
          tmp_input_value = (input_value != NULL) ? input_value : target_option->missing_value.string_value;
          if (tmp_input_value != NULL)
            {
            found_match = 0;

            if ((*ptr.string_array_ptr) != NULL)
              for (i = 0; (*ptr.string_array_ptr)[i] != NULL; i++)
                if (strcmp((*ptr.string_array_ptr)[i], tmp_input_value) == 0)
                  {
                  found_match = 1;
                  break;
                  }

            if (!found_match)
              {
              if (append_string(current_settings, ptr.string_array_ptr, tmp_input_value, strlen(tmp_input_value)))
                {
                changed_value = 1;

                if (current_settings->filter_grace < target_option->set_grace)
                  current_settings->filter_grace = target_option->set_grace;

                if (return_value < target_option->set_consequence)
                  return_value = target_option->set_consequence;
                }
              else
                return_value = FILTER_DECISION_ERROR;
              }

            if (changed_value &&
                (target_option->value_type == CONFIG_TYPE_COMMAND_ARRAY))
              {
              for (i = 0; (*ptr.string_array_ptr)[i] != NULL; i++);
              if (i > 0)
                for (j = 0; (*ptr.string_array_ptr)[i - 1][j] != '\0'; j++)
                  if ((*ptr.string_array_ptr)[i - 1][j] == COMMAND_LINE_SPACER)
                    (*ptr.string_array_ptr)[i - 1][j] = ' ';
              }
            }
          else
            {
            usage(current_settings, 0, ERROR_CONFIG_BAD_VALUE, target_option->getopt_option.name, input_value);
            return_value = FILTER_DECISION_ERROR;
            }
          }
        else
          {
          usage(current_settings, 0, ERROR_CONFIG_UNKNOWN_OPTION, target_option->getopt_option.name);
          return_value = FILTER_DECISION_ERROR;
          }

        break;
      default:
        break;
      }

  if (changed_value &&
      (target_option->additional_set_actions != NULL))
    return_value = (*target_option->additional_set_actions)(current_settings, return_value, input_value, history);

  return(return_value);
  }

int process_config_file(struct filter_settings *current_settings, char *config_filename, int return_value, struct previous_action *history)
  {
  FILE *tmp_file;
  int line_num;
  int i;
  int strlen_directive;
  int strlen_value;
  char tmp_buf[MAX_FILE_BUF + 1];
  int strlen_buf;
  char directive[MAX_FILE_BUF + 1];
  char value[MAX_FILE_BUF + 1];
  struct previous_action tmp_action;
  struct previous_action *current_action;

  for (current_action = history; current_action != NULL; current_action = current_action->prev)
    if (!strcmp(config_filename, current_action->data))
      break;

  if (current_action == NULL)
    {
    tmp_action.data = config_filename;
    tmp_action.prev = history;
    current_action = &tmp_action;

    if ((tmp_file = fopen(config_filename, "r")) != NULL)
      {
      line_num = 0;

      while (!feof(tmp_file) &&
             (line_num < MAX_FILE_LINES))
        {
        if ((fscanf(tmp_file, "%" STRINGIFY(MAX_FILE_BUF) "[^\r\n]", tmp_buf) == 1) &&
            (tmp_buf[0] != '#') &&
            ((strlen_buf = strlen(tmp_buf)) > 0))
          {
          value[0] = '\0';

          if ((sscanf(tmp_buf, "%[^= \t\r\n]%*[ \t]=%*[ \t]%[^\r\n]", directive, value) != 2) &&
              (sscanf(tmp_buf, "%[^= \t\r\n]=%*[ \t]%[^\r\n]", directive, value) != 2) &&
              (sscanf(tmp_buf, "%[^= \t\r\n]%*[ \t]=%[^\r\n]", directive, value) != 2))
            sscanf(tmp_buf, "%[^= \t\r\n]=%[^\r\n]", directive, value);

          strlen_value = strlen(value);
          for (i = strlen_value - 1; i >= 0; i--)
            if (isspace((int)value[i]))
              value[i] = '\0';
            else
              break;

          for (strlen_directive = 0; directive[strlen_directive] != '\0'; strlen_directive++)
            directive[strlen_directive] = tolower((int)directive[strlen_directive]);

          for (i = 0; current_settings->option_list[i].value_type != CONFIG_TYPE_NONE; i++)
            if (!strncmp(directive, current_settings->option_list[i].getopt_option.name, strlen_directive))
              {
              return_value = set_config_value(current_settings, &current_settings->option_list[i], value, return_value, current_action);
              break;
              }

          if (current_settings->option_list[i].value_type == CONFIG_TYPE_NONE)
            {
            spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_UNKNOWN_CONFIG, config_filename, line_num, directive);
            return_value = FILTER_DECISION_ERROR;
            }
          }

        fscanf(tmp_file, "%*1[\r\n]");
        line_num++;
        }

      fclose(tmp_file);
      }
    else
      {
      spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_OPEN_CONFIG, config_filename, strerror(errno));
      return_value = FILTER_DECISION_ERROR;
      }
    }

  return(return_value);
  }

int process_command_line(struct filter_settings *current_settings, int argc, char *argv[], char *envp[])
  {
  int return_value;
  int i;
  int tmp_argc;
  int opt;
  int displayed_banner;

  return_value = FILTER_DECISION_UNDECIDED;
  opterr = 0;
  displayed_banner = 0;

  current_settings->server_ip = find_environment_variable(envp, ENVIRONMENT_REMOTE_IP, STRLEN(ENVIRONMENT_REMOTE_IP));
  if ((current_settings->server_ip != NULL) &&
      (current_settings->server_ip[0] != '\0'))
    {
    current_settings->strlen_server_ip = (current_settings->server_ip != NULL) ? strlen(current_settings->server_ip) : 0;
    dns_ptr(current_settings->server_ip, current_settings);
    }
  else
    {
    current_settings->server_ip = DEFAULT_REMOTE_IP;
    current_settings->strlen_server_ip = STRLEN(DEFAULT_REMOTE_IP);
    current_settings->server_name[0] = '\0';
    current_settings->strlen_server_name = 0;
    }

  tmp_argc = 1;
  while ((return_value != FILTER_DECISION_ERROR) &&
         ((opt = getopt_long(argc, argv, current_settings->short_options, current_settings->long_options, NULL)) != -1))
    {
    if ((opt != '?') &&
        (opt != ':'))
      {
      for (i = 0; current_settings->option_list[i].value_type != CONFIG_TYPE_NONE; i++)
        if (current_settings->option_list[i].getopt_option.val == opt)
          {
          return_value = set_config_value(current_settings, &current_settings->option_list[i], optarg, return_value, NULL);
          break;
          }

      if (current_settings->option_list[i].value_type == CONFIG_TYPE_NONE)
        {
        usage(current_settings, 0, ERROR_CONFIG_UNKNOWN_OPTION, argv[tmp_argc]);
        return_value = FILTER_DECISION_ERROR;
        }
      }
    else
      {
      usage(current_settings, 0, ERROR_CONFIG_UNKNOWN_OPTION, argv[tmp_argc]);
      return_value = FILTER_DECISION_ERROR;
      }

    tmp_argc++;
    }

  if (return_value == FILTER_DECISION_ERROR)
    exit(0);
  else if (return_value == FILTER_DECISION_CONFIG_TEST)
    {
    test_settings(current_settings);
    exit(0);
    }
  else if (optind < argc)
    current_settings->child_argv = argv + optind;
  else
    {
    usage(current_settings, 0, ERROR_CONFIG_NO_COMMAND, NULL);
    exit(0);
    }

  return(return_value);
  }
