/*
  spamdyke -- a filter for stopping spam at connection time.
  Copyright (C) 2007 Sam Clippinger (samc (at) silence (dot) org)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 as
  published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <sys/types.h>
#include <string.h>
#include <syslog.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <ctype.h>
#include <netdb.h>
#include "spamdyke.h"
#include "log.h"
#include "dns.h"

#ifdef HAVE_NAMESER_COMPAT
#include <arpa/nameser_compat.h>
#endif /* HAVE_NAMESER_COMPAT */

/*
 * The DNS packet format is not well documented outside of the RFCs.  There is
 * almost no sample or tutorial code to be found on the internet outside of the
 * sendmail source code, which is pretty hard to read.
 *
 * Basically, each DNS packet starts with a HEADER structure, defined in
 * arpa/nameser.h (or arpa/nameser_compat.h on Linux).  Most of the time, the
 * header can be skipped.
 *
 * After the header, the nameserver returns all of the "questions" it was asked,
 * so the answers will make sense.  If you're asking more than one "question"
 * per query, this is important.  Otherwise, skip them by finding the size of
 * each question with dn_skipname() and advancing past them.  The number of
 * questions is found in the qdcount field of the header.
 *
 * Next is the answer section, which can contain many answers, though multiple
 * answers may not make much sense for all query types.  The number of answers
 * is found in the ancount field of the header.  Within each answer, the first
 * field is the name that was queried, for reference.  It can be skipped with
 * dn_skipname().
 *
 * After that comes the type in a 16 bit field, then the class in a 16 bit
 * field, then the time-to-live (TTL) in a 32 bit field, then the answer size
 * in a 16 bit field.  The type and size are important; the class and the ttl
 * can usually be ignored.  The format of the rest of the answer field is
 * different depending on the type.
 *
 * IF THE TYPE IS A:
 * The first 4 bytes are the four octets of the IP address.
 *
 * IF THE TYPE IS T_TXT:
 * The first 8 bits are an unsigned integer indicating the total length of
 * the text response.  The following bytes are the ASCII text of the response.
 *
 * IF THE TYPE IS T_PTR OR T_NS:
 * All of the bytes are the compressed name of the result.  They can be
 * decoded with dn_expand().
 *
 * IF THE TYPE IS T_CNAME:
 * All of the bytes are the compressed name of the CNAME entry.  They can be
 * decoded with dn_expand().
 *
 * IF THE TYPE IS T_MX:
 * Each answer begins with an unsigned 16 bit integer indicating the preference
 * of the mail server (lower preferences should be contacted first).  The
 * remainder of the answer is the mail server name.  It can be decoded with
 * dn_expand().
 *
 * IF THE TYPE IS T_SOA:
 * The first section of bytes are the compressed name of the primary NS server.
 * They can be decoded with dn_expand().  The second section of bytes are the
 * compressed name of the administrator's mailbox.  They can be decoded with
 * dn_expand().  After the end of the mailbox data, five 32 bit integers give
 * the serial number, the refresh interval, the retry interval, the expiration
 * limit and the minimum time to live, in that order.
 *
 * SEE ALSO:
 *   RFC 1035
 *   http://www.zytrax.com/books/dns/ch15/
 */
int dns_txt(char *target_name, struct filter_settings *current_settings, int set_rejection, struct previous_action *history)
  {
  static int initialized = 0;
  int return_value;
  int i;
  unsigned char answer[PACKETSZ];
  unsigned char host[128];
  unsigned char *answer_ptr;
  int answer_length;
  int size;
  int txt_length;
  int type;
  int num_questions;
  int num_answers;
  struct previous_action current_lookup;
  struct previous_action *tmp_lookup;

  return_value = 0;

  if (!initialized)
    {
    if (res_init() >= 0)
      initialized = 1;
    else
      spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_RES_INIT);
    }

  if (initialized)
    {
    memset(answer, 0, PACKETSZ);

    if ((target_name != NULL) &&
        (answer_length = res_search(target_name, C_IN, T_TXT, answer, PACKETSZ)) >= 0)
      {
      /* Skip DNS packet header */
      answer_ptr = answer + sizeof(HEADER);

      /* Skip the questions */
      num_questions = ntohs((unsigned short)((HEADER *)&answer)->qdcount);
      for (i = 0; i < num_questions; i++)
        if ((size = dn_skipname(answer_ptr, answer + answer_length)) >= 0)
          answer_ptr += size + QFIXEDSZ;
        else
          break;

      if (i == num_questions)
        {
        num_answers = ntohs((unsigned short)((HEADER *)&answer)->ancount);
        for (i = 0; (i < num_answers) && !return_value; i++)
          if ((size = dn_skipname(answer_ptr, answer + answer_length)) >= 0)
            {
            answer_ptr += size;
            GETSHORT(type, answer_ptr);
            answer_ptr += INT16SZ; /* class */
            answer_ptr += INT32SZ; /* ttl */
            GETSHORT(size, answer_ptr);

            if (type == T_TXT)
              {
              txt_length = (unsigned char)*answer_ptr;
              answer_ptr++;

              if ((txt_length >= 0) &&
                  (txt_length < size))
                {
                if (set_rejection)
                  {
                  memcpy(&current_settings->rejection_buf, &rejection[REJECTION_RBL], sizeof(struct rejection_data));
                  current_settings->rejection_buf.strlen_reject_message = snprintf(current_settings->reject_message_buf, MAX_BUF, "%.*s", MINVAL(txt_length, MAX_BUF - 2), answer_ptr);
                  current_settings->rejection_buf.reject_message = current_settings->reject_message_buf;
                  current_settings->rejection = &current_settings->rejection_buf;
                  }

                return_value = 1;
                answer_ptr += size - 1;
                }
              else
                {
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_DNS_RESPONSE, target_name);
                break;
                }
              }
            else if (type == T_CNAME)
              {
              if ((size = dn_expand(answer, answer + answer_length, answer_ptr, (char *)host, sizeof(host))) >= 0)
                {
                current_lookup.data = target_name;
                current_lookup.prev = history;

                tmp_lookup = &current_lookup;
                while (tmp_lookup != NULL)
                  if (strcasecmp((char *)host, tmp_lookup->data) == 0)
                    break;
                  else
                    tmp_lookup = tmp_lookup->prev;

                return_value = (tmp_lookup == NULL) ? dns_txt((char *)host, current_settings, set_rejection, &current_lookup) : 0;
                answer_ptr += size;
                }
              else
                {
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_DN_EXPAND, target_name);
                break;
                }
              }
            else
              {
              spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_TYPE_TXT, target_name);
              break;
              }
            }
          else
            {
            spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_DN_EXPAND, target_name);
            break;
            }
        }
      else
        spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_DN_SKIPNAME, target_name);
      }
    }

  return(return_value);
  }

int dns_ptr_lookup(char *target_name, struct filter_settings *current_settings, struct previous_action *history)
  {
  static int initialized = 0;
  int return_value;
  int i;
  unsigned char answer[PACKETSZ];
  unsigned char host[128];
  unsigned char *answer_ptr;
  unsigned char *tmp_ptr;
  int answer_length;
  int size;
  int txt_length;
  int type;
  int num_questions;
  int num_answers;
  struct previous_action current_lookup;
  struct previous_action *tmp_lookup;

  return_value = 0;

  if (!initialized)
    {
    if (res_init() >= 0)
      initialized = 1;
    else
      spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_RES_INIT);
    }

  if (initialized)
    {
    memset(answer, 0, PACKETSZ);

    if ((target_name != NULL) &&
        ((answer_length = res_search(target_name, C_IN, T_PTR, answer, PACKETSZ)) >= 0))
      {
      answer_ptr = answer + sizeof(HEADER);

      /* Skip the questions */
      num_questions = ntohs((unsigned short)((HEADER *)&answer)->qdcount);
      for (i = 0; i < num_questions; i++)
        if ((size = dn_skipname(answer_ptr, answer + answer_length)) >= 0)
          answer_ptr += size + QFIXEDSZ;
        else
          break;

      if (i == num_questions)
        {
        num_answers = ntohs((unsigned short)((HEADER *)&answer)->ancount);
        for (i = 0; (i < num_answers) && !return_value; i++)
          if ((size = dn_skipname(answer_ptr, answer + answer_length)) >= 0)
            {
            answer_ptr += size;
            GETSHORT(type, answer_ptr);
            answer_ptr += INT16SZ; /* class */
            answer_ptr += INT32SZ; /* ttl */
            GETSHORT(size, answer_ptr);

            if (type == T_PTR)
              {
              if (size > 2)
                {
                tmp_ptr = answer_ptr;
                while (((tmp_ptr - answer_ptr) < size) &&
                       ((txt_length = tmp_ptr[0]) > 0))
                  {
                  tmp_ptr[0] = '.';
                  tmp_ptr += txt_length + 1;
                  }

                current_settings->strlen_server_name = MINVAL(MAX_BUF, size - 2);
                for (i = 0; i < current_settings->strlen_server_name; i++)
                  current_settings->server_name[i] = tolower((int)(answer_ptr + 1)[i]);

                return_value = 1;
                answer_ptr += size;
                }
              else
                {
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_DNS_RESPONSE, target_name);
                break;
                }
              }
            else if (type == T_CNAME)
              {
              if ((size = dn_expand(answer, answer + answer_length, answer_ptr, (char *)host, sizeof(host))) >= 0)
                {
                current_lookup.data = target_name;
                current_lookup.prev = history;

                tmp_lookup = &current_lookup;
                while (tmp_lookup != NULL)
                  if (strcasecmp((char *)host, tmp_lookup->data) == 0)
                    break;
                  else
                    tmp_lookup = tmp_lookup->prev;

                return_value = (tmp_lookup == NULL) ? dns_ptr_lookup((char *)host, current_settings, &current_lookup) : 0;
                answer_ptr += size;
                }
              else
                {
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_DN_EXPAND, target_name);
                break;
                }
              }
            else
              {
              spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_TYPE_PTR, target_name);
              break;
              }
            }
          else
            {
            spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_DN_SKIPNAME, target_name);
            break;
            }
        }
      else
        spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_DN_SKIPNAME, target_name);
      }
    }

  return(return_value);
  }

int dns_ptr(char *target_ip, struct filter_settings *current_settings)
  {
  int return_value;
  char ip_octets[4][4];
  char rdns_name[MAX_RDNS + 1];

  return_value = 0;

  if ((target_ip != NULL) &&
      (sscanf(target_ip, "%3[0-9].%3[0-9].%3[0-9].%3[0-9]", ip_octets[0], ip_octets[1], ip_octets[2], ip_octets[3]) == 4) &&
      snprintf(rdns_name, MAX_RDNS, "%s.%s.%s.%s%s.", ip_octets[3], ip_octets[2], ip_octets[1], ip_octets[0], RDNS_SUFFIX))
    return_value = dns_ptr_lookup(rdns_name, current_settings, NULL);

  return(return_value);
  }

int dns_mx(struct filter_settings *current_settings, char *target_name, int strlen_target_name, struct previous_action *history)
  {
  static int initialized = 0;
  int return_value;
  int i;
  int j;
  unsigned char answer[PACKETSZ];
  unsigned char host[128];
  char new_target[MAX_BUF + 1];
  unsigned char *answer_ptr;
  int strlen_new_target;
  int answer_length;
  int size;
  int type;
  int num_questions;
  int num_answers;
  int preference;
  struct hostent *tmp_hostent;
  struct previous_action current_lookup;
  struct previous_action *tmp_lookup;

  return_value = 0;

  if (!initialized)
    {
    if (res_init() >= 0)
      initialized = 1;
    else
      spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_RES_INIT);
    }

  if (initialized)
    {
    memset(answer, 0, PACKETSZ);

    if ((target_name != NULL) &&
        (strlen_target_name > 0) &&
        ((strlen_new_target = snprintf(new_target, MAX_BUF, "%.*s.", strlen_target_name, target_name)) > 0) &&
        (strlen_new_target < MAX_BUF))
      {
      if (gethostbyname(new_target) != NULL)
        return_value = 1;
      else if ((answer_length = res_search(new_target, C_IN, T_MX, answer, PACKETSZ)) >= 0)
        {
        answer_ptr = answer + sizeof(HEADER);

        /* Skip the questions */
        num_questions = ntohs((unsigned short)((HEADER *)&answer)->qdcount);
        for (i = 0; i < num_questions; i++)
          if ((size = dn_skipname(answer_ptr, answer + answer_length)) >= 0)
            answer_ptr += size + QFIXEDSZ;
          else
            break;

        if (i == num_questions)
          {
          num_answers = ntohs((unsigned short)((HEADER *)&answer)->ancount);
          for (i = 0; (i < num_answers) && !return_value; i++)
            if ((size = dn_skipname(answer_ptr, answer + answer_length)) >= 0)
              {
              answer_ptr += size;
              GETSHORT(type, answer_ptr);
              answer_ptr += INT16SZ; /* class */
              answer_ptr += INT32SZ; /* ttl */
              GETSHORT(size, answer_ptr);

              if (type == T_MX)
                {
                GETSHORT(preference, answer_ptr);
                if ((size = dn_expand(answer, answer + answer_length, answer_ptr, (char *)host, sizeof(host))) >= 0)
                  {
                  if ((tmp_hostent = gethostbyname((char *)host)) != NULL)
                    for (j = 0; tmp_hostent->h_addr_list[j] != NULL; j++)
                      if (((struct in_addr *)tmp_hostent->h_addr_list[j])->s_addr != INADDR_LOOPBACK)
                        {
                        return_value = 1;
                        break;
                        }

                  answer_ptr += size;
                  }
                else
                  {
                  spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_DN_EXPAND, new_target);
                  break;
                  }
                }
              else if (type == T_CNAME)
                {
                if ((size = dn_expand(answer, answer + answer_length, answer_ptr, (char *)host, sizeof(host))) >= 0)
                  {
                  current_lookup.data = target_name;
                  current_lookup.prev = history;

                  tmp_lookup = &current_lookup;
                  while (tmp_lookup != NULL)
                    if (strcasecmp((char *)host, tmp_lookup->data) == 0)
                      break;
                    else
                      tmp_lookup = tmp_lookup->prev;

                  return_value = (tmp_lookup == NULL) ? dns_mx(current_settings, (char *)host, strlen((char *)host), &current_lookup) : 0;
                  answer_ptr += size;
                  }
                else
                  {
                  spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_DN_EXPAND, new_target);
                  break;
                  }
                }
              else
                {
                spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_TYPE_MX, new_target);
                break;
                }
              }
            else
              {
              spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_DN_SKIPNAME, new_target);
              break;
              }
          }
        else
          spamdyke_log(current_settings, LOG_LEVEL_ERROR, LOG_ERROR_DN_SKIPNAME, new_target);
        }
      }
    }

  return(return_value);
  }

int check_dnsrbl(struct filter_settings *current_settings, char *target_ip_address, char *target_dnsrbl, int set_rejection)
  {
  int return_value;
  char rbl_name[MAX_BUF + 1];
  char ip_octets[4][4];

  return_value = 0;

  if ((target_ip_address != NULL) &&
      (target_dnsrbl != NULL) &&
      (sscanf(target_ip_address, "%3[0-9].%3[0-9].%3[0-9].%3[0-9]", ip_octets[0], ip_octets[1], ip_octets[2], ip_octets[3]) == 4) &&
      snprintf(rbl_name, MAX_BUF, "%s.%s.%s.%s.%s.", ip_octets[3], ip_octets[2], ip_octets[1], ip_octets[0], target_dnsrbl))
    {
    if (dns_txt(rbl_name, current_settings, set_rejection, NULL))
      return_value = 1;
    else if (gethostbyname(rbl_name) != NULL)
      {
      if (set_rejection)
        {
        memcpy(&current_settings->rejection_buf, &rejection[REJECTION_RBL], sizeof(struct rejection_data));
        current_settings->rejection_buf.strlen_reject_message = snprintf(current_settings->reject_message_buf, MAX_BUF, rejection[REJECTION_RBL].reject_message, target_dnsrbl);
        current_settings->rejection_buf.reject_message = current_settings->reject_message_buf;
        current_settings->rejection = &current_settings->rejection_buf;
        }

      return_value = 1;
      }
    }

  return(return_value);
  }

int check_rhsbl(struct filter_settings *current_settings, char *target_right_side, char *target_rhsbl, int set_rejection)
  {
  int return_value;
  char rbl_name[MAX_BUF + 1];
  char *tmp_ptr;
  int strlen_right_side;
  char *domain_start[2];

  return_value = 0;

  if ((target_right_side != NULL) &&
      (target_rhsbl != NULL))
    {
    domain_start[0] = NULL;
    domain_start[1] = NULL;

    strlen_right_side = strlen(target_right_side);
    for (tmp_ptr = (target_right_side + strlen_right_side - 1); (tmp_ptr != NULL) && (tmp_ptr > target_right_side); tmp_ptr--)
      if (tmp_ptr[0] == '.')
        {
        if (domain_start[0] == NULL)
          domain_start[0] = tmp_ptr + 1;
        else if (domain_start[1] == NULL)
          {
          domain_start[1] = tmp_ptr + 1;
          break;
          }
        }

    if ((tmp_ptr == target_right_side) &&
        (tmp_ptr != domain_start[0]))
      domain_start[1] = target_right_side;

    if (domain_start[1] != NULL)
      {
      tmp_ptr = target_right_side;

      while ((tmp_ptr != NULL) &&
             (tmp_ptr <= domain_start[1]))
        {
        if (snprintf(rbl_name, MAX_BUF, "%s.%s.", tmp_ptr, target_rhsbl))
          {
          if (dns_txt(rbl_name, current_settings, set_rejection, NULL))
            {
            return_value = 1;
            break;
            }
          else if (gethostbyname(rbl_name) != NULL)
            {
            if (set_rejection)
              {
              memcpy(&current_settings->rejection_buf, &rejection[REJECTION_RHSBL], sizeof(struct rejection_data));
              current_settings->rejection_buf.strlen_reject_message = snprintf(current_settings->reject_message_buf, MAX_BUF, rejection[REJECTION_RHSBL].reject_message, target_rhsbl);
              current_settings->rejection_buf.reject_message = current_settings->reject_message_buf;
              current_settings->rejection = &current_settings->rejection_buf;
              }

            return_value = 1;
            break;
            }
          }

        if ((tmp_ptr = strchr(tmp_ptr, '.')) != NULL)
          tmp_ptr++;
        }
      }
    }

  return(return_value);
  }
